from .__version__ import __version__  # noqa: F401
from .model import Model  # noqa: F401
from .muscle import Muscle  # noqa: F401
from .muscle import PotvinFuglevandMuscle  # noqa: F401
from .muscle import StandardMuscle  # noqa: F401
from .potvin_fuglevand_2017_muscle_fibers import PotvinFuglevand2017MuscleFibers  # noqa: F401
from .potvin_fuglevand_2017_motor_neuron_pool import PotvinFuglevand2017MotorNeuronPool  # noqa: F401
from .pymuscle_fibers import PyMuscleFibers  # noqa: F401
from .hill_type import (
    contractile_element_force_length_curve,
    contractile_element_force_velocity_curve
)



import numpy as np
import colorlover as cl
from numpy import ndarray
from plotly.offline import plot


class PotvinChart(object):

    def __init__(self, time_by_forces: ndarray, step_size: float):

        forces_by_time = np.array(time_by_forces).T
        motor_unit_count, steps = forces_by_time.shape
        sim_duration = steps * step_size
        times = np.arange(0.0, sim_duration, step_size)

        # Setting colors for plot.
        potvin_scheme = [
            'rgb(115, 0, 0)',
            'rgb(252, 33, 23)',
            'rgb(230, 185, 43)',
            'rgb(107, 211, 100)',
            'rgb(52, 211, 240)',
            'rgb(36, 81, 252)',
            'rgb(0, 6, 130)'
        ]

        # It's hacky but also sorta cool.
        c = cl.to_rgb(cl.interp(potvin_scheme, motor_unit_count))
        c = [val.replace('rgb', 'rgba') for val in c]
        c = [val.replace(')', ',{})') for val in c]

        # Assing non-public attributes
        self._step_size = step_size
        self._forces_by_time = forces_by_time
        self._c = c
        self._times = times

        # Assign public attributes
        self.motor_unit_count = motor_unit_count

    def _get_color(self, trace_index: int) -> str:
        # The first and every 20th trace should be full opacity
        alpha = 0.2
        if trace_index == 0 or ((trace_index + 1) % 20 == 0):
            alpha = 1.0
        color = self._c[trace_index].format(alpha)
        return color

    def display(self) -> None:
        # Per Motor Unit Force
        data = []
        for i, t in enumerate(self._forces_by_time):
            trace = dict(
                x=self._times,
                y=t,
                name=i + 1,
                marker=dict(
                    color=self._get_color(i)
                ),
            )
            data.append(trace)

        layout = dict(
            title='Motor Unit Forces by Time',
            yaxis=dict(
                title='Motor unit force (relative to MU1 tetanus)'
            ),
            xaxis=dict(
                title='Time (s)'
            )
        )

        fig = dict(
            data=data,
            layout=layout
        )
        plot(fig, filename='forces-by-time.html', validate=False)



from numpy import ndarray

from .potvin_fuglevand_2017_muscle_fibers import PotvinFuglevand2017MuscleFibers


class PyMuscleFibers(PotvinFuglevand2017MuscleFibers):
    """
    Encapsulates the muscle fibers portions of the motor unit model. Currently
    a thin wrapper to implement fiber recovery.

    This is the standard muscle fiber model for PyMuscle. Which underlying
    model it extends is expected to change over time. This class exists to
    support the needs of users in the machine learning community. Users who
    belong to the physiology, biomechanical, or medical communities should
    use one of the base classes (such as PotvinFuglevand2017MuscleFibers) as
    those are strict implementations of published work.

    :param motor_unit_count: Number of motor units in the muscle
    :param force_conversion_factor: The ratio of newtons to arbitrary force
        units. All peak twitch forces are calculated internally to lie in a
        range of 0 to 100 arbitrary force units. The maximum force these
        fibers can theoretically produce is the sum of those peak twitch forces.
        To relate the arbitrary force units to SI units you need to provide
        a conversion factor. Increasing this value is essentially saying that
        a given motor unit produces more force than the default value would
        suggest.
    :param max_twitch_amplitude: Max twitch force within the pool
    :param max_contraction_time:
        [milliseconds] Maximum contraction time for a motor unit
    :param contraction_time_range:
        The scale between the fastest contraction time and the slowest
    :fatigue_factor_first_unit:
        The nominal fatigability of the first motor unit in percent / second
    :fatigability_range:
        The scale between the fatigability of the first motor unit and the last
    :contraction_time_change_ratio:
        For each percent of force lost during fatigue, what percentage should
        contraction increase?

    .. todo::
        The argument naming isn't consistent. Sometimes we use 'max' and other
        times we use 'last unit'. Can these be made consistent?

    Usage::

      from pymuscle import PyMuscleFibers as Fibers

      motor_unit_count = 60
      fibers = Fibers(motor_unit_count)
      motor_neuron_firing_rates = np.rand(motor_unit_count) * 10.0
      step_size = 0.01
      force = fibers.step(motor_neuron_firing_rates, step_size)
    """
    def __init__(
        self,
        *args,
        force_conversion_factor: float=0.028,
        **kwargs
    ):
        super().__init__(*args, **kwargs)

        # Ratio of newtons (N) to internal arbitrary force units
        self.force_conversion_factor = force_conversion_factor

        # Define recovery rates
        # Averaged from data in Liu et al. 2002, Table 2
        max_recovery_rate = self._max_fatigue_rate / 2.53
        # Recovery should ~= fatigue for small units, <= for medium units and
        # << for largest units
        # Re-uses the same method as calculating fatigabilities.
        recovery_range = max_recovery_rate / self._nominal_fatigabilities[0]
        self._recovery_rates = self._calc_nominal_fatigabilities(
            self.motor_unit_count,
            recovery_range,
            max_recovery_rate,
            self._peak_twitch_forces
        )

    def _update_fatigue(
        self,
        normalized_forces: ndarray,
        step_size: float
    ) -> None:
        """
        Updates current twitch forces and contraction times. This overrides
        the parent method to add in recovery calculations.

        :param normalized_forces:
            Array of scaled forces. Used to weight how much fatigue will be
            generated in this step.
        :param step_size: How far time has advanced in this step.
        """
        fatigues = (self._nominal_fatigabilities * normalized_forces) * step_size
        self._current_peak_forces -= fatigues

        # Apply recovery for units producing no force
        self._apply_recovery(normalized_forces, step_size)

        # Zero out negative values
        self._current_peak_forces[self._current_peak_forces < 0] = 0.0

        # Clip max values
        over = self._current_peak_forces > self._peak_twitch_forces
        self._current_peak_forces[over] = self._peak_twitch_forces[over]

        # Apply fatigue to contraction times
        self._update_contraction_times()

    def _apply_recovery(
        self,
        normalized_forces: ndarray,
        step_size: float
    ) -> None:
        """
        Apply recovery to motor units not producing force in this step.

        TODO - Finalize the strategy used below
        """
        recovering = normalized_forces <= 0

        # Strategy 1 - Linear recovery at fatigue rates
        # recovery = self._nominal_fatigabilities[recovering] * step_size

        # Strategy 2 - Uses inverse of fatigue rates in an asymptotic approach
        # peak = self._peak_twitch_forces[recovering]
        # current = self._current_peak_forces[recovering]
        # recovery_ratio = (peak - current) / peak
        # recovery = (self._nominal_fatigabilities[recovering] * recovery_ratio) * step_size

        # Strategy 3 - Uses linear approach with calculated recovery rates
        # recovery = self._recovery_rates[recovering] * step_size

        # Strategy 4 - Combine 2 and 3
        peak = self._peak_twitch_forces[recovering]
        current = self._current_peak_forces[recovering]
        recovery_ratio = (peak - current) / peak
        recovery = (self._recovery_rates[recovering] * recovery_ratio) * step_size

        self._current_peak_forces[recovering] += recovery



from numpy import ndarray


class Model(object):
    """
    Base model class from which other models should inherit
    """
    def __init__(
        self,
        motor_unit_count: int
    ):
        self.motor_unit_count = motor_unit_count

    def step(self, inputs: ndarray, step_size: float):
        """
        Child classes must implement this method.
        """
        raise NotImplementedError



"""
Contains base Muscle class and its immediate descendants.
"""

import numpy as np
from typing import Union

from .potvin_fuglevand_2017_muscle_fibers import PotvinFuglevand2017MuscleFibers
from .potvin_fuglevand_2017_motor_neuron_pool import PotvinFuglevand2017MotorNeuronPool
from .pymuscle_fibers import PyMuscleFibers
from .model import Model


class Muscle(object):
    """
    A user-created :class:`Muscle <Muscle>` object.

    Used to simulate the input-output relationship between motor neuron
    excitation and muscle fibers contractile state over time.

    :param motor_neuron_pool_model:
        The motor neuron pool implementation to use with this muscle.
    :param muscle_fibers_model:
        The muscle fibers model implementation to use with this muscle.
    :param motor_unit_count: How many motor units comprise this muscle.

    Usage::

        from pymuscle import (Muscle,
                              PotvinFuglevand2017MotorNeuronPool as Pool,
                              PotvinFuglevand2017MuscleFibers as Fibers)

        motor_unit_count = 60
        muscle = Muscle(
            Pool(motor_unit_count),
            Fibers(motor_unit_count),
        )
        excitation = 32.0
        force = muscle.step(excitation, 1 / 50.0)
    """

    def __init__(
        self,
        motor_neuron_pool_model: Model,
        muscle_fibers_model: Model,
    ):
        assert motor_neuron_pool_model.motor_unit_count == \
            muscle_fibers_model.motor_unit_count

        self._pool = motor_neuron_pool_model
        self._fibers = muscle_fibers_model

    @property
    def motor_unit_count(self):
        return self._pool.motor_unit_count

    @property
    def max_excitation(self):
        return self._pool.max_excitation

    @property
    def current_forces(self):
        return self._fibers.current_forces

    def step(
        self,
        motor_pool_input: Union[float, np.ndarray],
        step_size: float
    ) -> float:
        """
        Advances the muscle model one step.

        :param motor_pool_input:
            Either a single value or an array of values representing the
            excitatory input to the motor neuron pool for this muscle
        :param step_size:
            How far to advance the simulation in time for this step.
        """

        # Expand a single input to the muscle to a full array
        if isinstance(motor_pool_input, float) or \
           isinstance(motor_pool_input, int):
            motor_pool_input = np.full(
                self._pool.motor_unit_count,
                motor_pool_input
            )

        motor_pool_output = self._pool.step(motor_pool_input, step_size)
        return self._fibers.step(motor_pool_output)


class PotvinFuglevandMuscle(Muscle):
    """
    A thin wrapper around :class:`Muscle <Muscle>` which pre-selects the
    Potvin fiber and motor neuron models.
    """

    def __init__(
        self,
        motor_unit_count: int,
        apply_central_fatigue: bool = True,
        apply_peripheral_fatigue: bool = True,
        pre_calc_firing_rates: bool = False
    ):
        pool = PotvinFuglevand2017MotorNeuronPool(
            motor_unit_count,
            apply_fatigue=apply_central_fatigue,
            pre_calc_firing_rates=pre_calc_firing_rates
        )
        fibers = PotvinFuglevand2017MuscleFibers(
            motor_unit_count,
            apply_fatigue=apply_peripheral_fatigue
        )

        super().__init__(
            motor_neuron_pool_model=pool,
            muscle_fibers_model=fibers
        )


class StandardMuscle(Muscle):
    """
    A wrapper around :class:`Muscle <Muscle>` which pre-selects the
    Potvin motor neuron model and the PyMuscle specific fiber model.

    The API for this class is oriented toward use along side physics sims so
    you primarily specify the max_force (in newtons) you want the muscle to
    output as opposed to motor_unit_count as with other classes in this
    library.

    It is expected that this will use a motor neuron model specific to PyMuscle
    (to be called the PyMuscleMotorNeuronPool) in the future.

    This muscle does *not* include central (motor neuron) fatigue as the
    equations for recovery are not yet available.

    This muscle does include both *peripheral* fatigue and recovery.

    :param max_force: Maximum voluntary isometric force this muscle can produce
        when fully rested and at maximum excitation. (Newtons)
        This along with the force_conversion_factor will determine the number
        of simulated motor units.
    :param force_conversion_factor: This library uses biological defaults to
        convert the desired max_force into a number of motor units which
        make up a muscle. This can result in a large number of motor units
        which may be slow. To improve performance (but diverge from biology)
        you can change this force conversion factor.

        Note: It is likely the default value here will change with major
        versions as better biological data is found.
    """
    def __init__(
        self,
        max_force: float = 32.0,
        force_conversion_factor: float = 0.0123,
        apply_central_fatigue: bool = False,
        apply_peripheral_fatigue: bool = True,
        pre_calc_firing_rates: bool = False
    ):

        # Maximum voluntary isometric force this muscle will be able to produce
        self.max_force = max_force

        # Ratio of newtons (N) to internal arbitrary force units
        self.force_conversion_factor = force_conversion_factor

        # Max output in arbitrary units
        self.max_output = max_force / force_conversion_factor

        motor_unit_count = self.force_to_motor_unit_count(
            self.max_force,
            self.force_conversion_factor,
        )

        pool = PotvinFuglevand2017MotorNeuronPool(
            motor_unit_count,
            apply_fatigue=apply_central_fatigue,
            pre_calc_firing_rates=pre_calc_firing_rates
        )
        fibers = PyMuscleFibers(
            motor_unit_count,
            force_conversion_factor=force_conversion_factor,
            apply_fatigue=apply_peripheral_fatigue
        )

        super().__init__(
            motor_neuron_pool_model=pool,
            muscle_fibers_model=fibers
        )

    @staticmethod
    def force_to_motor_unit_count(
        max_force: float,
        conversion_factor: float
    ) -> int:
        # This takes the relationship between force production
        # and number of motor units from Fuglevand 93 and solves
        # for motor units given desired force.

        # The reference muscle is the first dorsal interossei
        # https://en.wikipedia.org/wiki/Dorsal_interossei_of_the_hand

        # The number of motor units for the FDI was estimated by
        # Feinstein et al. (1955) at 119.
        # https://www.ncbi.nlm.nih.gov/pubmed/14349537
        #
        # Caveats:
        #  - This estimate was from 1 dissection of an adult male
        #  - This esimate counted 'large nerve fibers' and then assumed
        #    40% of them would be afferent.

        # The Fuglevand experiments use 120 motor units.  The peak twitch
        # forces are calculated in the range of 0-100 arbitrary force units
        # with each motor unit being assigned a peak twitch force according
        # to the exponential function outlined in Fuglevand 93.
        #
        # The total possible force units this muscle could produce, after
        # assignment of peak twitch forces, is the sum of those peak twitch
        # forces or ~2609.03 arbitrary units.

        # The maximum voluntary force (MVC) for the FDI was taken from
        # Jahanmir-Zezhad et al.
        # https://www.ncbi.nlm.nih.gov/pmc/articles/PMC5450824/
        # MVC = 32 Newtons (N)

        # Thus if a modeled FDI with 120 motor units produces an MVC of
        # 2,216 units and a real FDI produces an MVC of 32N we can
        # relate them as follows
        #
        # conversion_factor = 32N / 2609.03 units = 0.01226509469 N/unit

        # See: https://www.desmos.com/calculator/b9xzsaqs1g
        # m = max_force
        # c = conversion_factor
        # u = motor unit count
        # u = 1 / ln( ((1-m/c) / (e^4.6 - m/c)))^(1/4.6) )
        r = (max_force / conversion_factor)
        n2 = 1 - r
        d2 = np.exp(4.6) - r
        inner = np.power((n2 / d2), (1 / 4.6))
        d = np.log(inner)  # This is natural log by default (ln)
        muf = 1 / d
        muc = int(np.ceil(muf))

        return muc

    def get_central_fatigue(self):
        raise NotImplementedError

    def get_peripheral_fatigue(self) -> float:
        """
        Returns fatigue level in the range 0.0 to 1.0 where:

        1.0 - Completely fatigued
        0.0 - Completely rested
        """
        fatigue = 1 - sum(self._fibers.current_peak_forces) / self.max_output
        return fatigue

    def step(
        self,
        motor_pool_input: Union[float, np.ndarray],
        step_size: float
    ) -> float:
        """
        Advances the muscle model one step.

        :param motor_pool_input:
            Either a single value or an array of values representing the
            excitatory input to the motor neuron pool for this muscle.
            Range is 0 - 1.0.
        :param step_size:
            How far to advance the simulation in time for this step.
        """

        # Rescale the input to the underlying rage for the motor pool
        motor_pool_input *= self.max_excitation
        return super().step(motor_pool_input, step_size)



"""
Implementations of equations used in Hill-Type muscle models

Not all equations for a traditional Hill-Type model are provided here. It is
expected that passive elements will be modeled by a physics simulator selected
by PyMuscle users.

Also it should be noted that these equations do not take pennetation angle
into account (which is common in Hill-type models). Users should check to
make sure their physics simulator is handling forces correctly to provide values
which vary by angle.
"""
import numpy as np


def contractile_element_force_length_curve(
    rest_length: float,
    current_length: float,
    curve_width_factor: float=17.33,
    peak_force_length: float=1.1
) -> float:
    """
    This normalizes length to the resting length of the tendon
    Anderson and others normalize by the length that would generate the
    most force. This is a gaussian-like relationship.

    :param rest_length: The resting length of the muscle
    :param current_length: The current length of the muscle
    :param curve_width_factor:
        A curve shape paramater which describes where force begins to approach
        0 for muscles shorter or longer than resting length.
        See https://www.desmos.com/calculator/qosweyyqfk for graph.
    :param peak_force_length:
        The ratio of the length of the muscle when it can generate maximum
        contractile force (sometimes known as the optimal length) and the
        resting length of the muscle.
    """
    norm_length = current_length / rest_length
    peak_force_length = 1.10  # Aubert 1951
    exponent = curve_width_factor * (-1 * (abs(norm_length - peak_force_length) ** 3))
    percentage_of_max_force = np.exp(exponent)
    return percentage_of_max_force


def contractile_element_force_velocity_curve(
    rest_length: float,
    current_length: float,
    prev_length: float,
    time_step: float,
    max_velocity: float = 3.0,
    max_exccentric_multiple: float = 1.8
) -> float:
    """
    Returns a value (0 < value < max_eccentric_multiple). This is a sigmoidal
    relationship.

    :param rest_length: The resting length of the muscle
    :param current_length: The current length of the muscle

    See https://www.desmos.com/calculator/gkcdsdcuyh for graph
    """
    velocity = ((prev_length - current_length) / rest_length) / time_step
    norm_velocity = velocity / max_velocity
    exponent = (0.04 - norm_velocity) / 0.18  # Values to make (0, 1)
    denominator = 1 + np.exp(exponent)
    force_multiple = max_exccentric_multiple - (max_exccentric_multiple / denominator)
    return force_multiple
